import queue
import sys
import threading
import time
import traceback
from typing import List, Optional, Union
from typeguard import typechecked
from azure.eventhub import EventHubProducerClient, EventData, EventDataBatch
from azure.eventhub._constants import JWT_TOKEN_SCOPE
from azure.identity import (
    DefaultAzureCredential,
    ManagedIdentityCredential,
)
from .log import (
    ERR_PREFIX, 
    LOGGING_QUEUE_CAPACITY,
    SINK_BATCH_MAX_IDLE_SECS,
    SINK_BATCH_MAX_SIZE,
    EventSink,
    Event,
    TerminateEvent
)


AzureCredentialTyping = Union[
    DefaultAzureCredential,
    ManagedIdentityCredential,
]


class EventHubSink(EventSink):
    @typechecked
    def __init__(
        self,
        name: str,
        hostname: Optional[str] = None,
        credential: Optional[AzureCredentialTyping] = None,
        conn_str: Optional[str] = None,
    ):
        '''
        Args:
            name: eventhub name
            hostname: Fully Qualified Namespace aka EH Endpoint URL (*.servicebus.windows.net)
            credential: token with access to eventhub
            conn_str: connection string of eventhub namespace

        Note: 
            either (hostname, credential) or conn_str you should provide
        '''
        # eventhub parameters
        if not conn_str and not (hostname and credential):
            raise Exception('neither (hostname, credential) nor conn_str is provided')
        self.name = name
        self.hostname = hostname
        self.credential = credential
        self.conn_str = conn_str
        self._verify_credential()

        # Critical: 
        #   EventHubSink is created in main process, but it's actually used in sub process. If you're reading 
        #   this piece of code, please keep in mind that all the variables set to `self` here will be copied to sub process. 
        #   Different references will exist in both processes at the same time, but they are not shared (unless it's a 
        #   multiprocess object like mp.Queue). 
        #   Again, __init__ funtion only set the initial value, and they will be copied to sub process. All the other 
        #   things are happend in sub process.
        #   Some variables have the problem that can't be serialized. So they will be assigned in `start` function, which 
        #   happenes in sub process.
        
        self.mutex = threading.RLock()
        self.terminated = threading.Event()
        self.q: queue.Queue[Event] = queue.Queue(maxsize=LOGGING_QUEUE_CAPACITY)
        ## This object (self.client) can only be created, consumed, and re-created in the same process where it will be used.
        self.client: Optional[EventHubProducerClient] = None
        self.loop_thread: Optional[threading.Thread] = None


    def __del__(self):
        self.close()


    def sink(self, evt: Event):
        try:
            self.q.put_nowait(evt)
        except queue.Full:
            msg = f'{ERR_PREFIX} EventHubSink shedding an event. Buff grew faster than stream could be written.'
            print(msg, file=sys.stderr)


    def start(self):
        with self.mutex:
            if self.loop_thread:
                return
            self.loop_thread = threading.Thread(target=self._loop_run, daemon=True)
            self.loop_thread.start()
    

    def close(self):
        with self.mutex:
            if not self.loop_thread:
                return
            if not self.terminated.is_set():
                self.terminated.set()
                try:
                    self.q.put_nowait(TerminateEvent())
                except queue.Full:
                    pass
        self.loop_thread.join()


    def _create_eventhub_client(self):
        if self.conn_str:
            return EventHubProducerClient.from_connection_string(self.conn_str, eventhub_name=self.name)
        elif self.hostname and self.credential:
            return EventHubProducerClient(self.hostname, self.name, self.credential)
        else:
            raise Exception('neither (hostname, credential) nor conn_str is provided')


    def _verify_credential(self):
        # case hostname and credential used for eventhub client
        if (not self.conn_str) and (self.hostname and self.credential):
            self.credential.get_token(JWT_TOKEN_SCOPE)


    def _require_eventhub_client(self):
        if not self.client:
            self.client = self._create_eventhub_client()


    def _reset_eventhub_client(self):
        self.client = None


    def _loop_run(self):
        while True:
            try:
                self._loop_run_inner()
            except Exception:
                traceback.print_exc()

            self._reset_eventhub_client()
            if self.terminated.is_set():
                break
            else:
                time.sleep(3)


    def _loop_run_inner(self):
        ''' condition to trigger a flush
            1. [buff size] >= SINK_BATCH_MAX_SIZE
            2. [now] - [last flush time] >= SINK_BATCH_MAX_IDLE_SECS
        '''
        buff: List[Event] = []
        last_flush_time = time.perf_counter()

        while not self.terminated.is_set():
            # get from q without block & check empty
            try:
                evt = self.q.get_nowait()
                if isinstance(evt, TerminateEvent):
                    continue
                else:
                    buff.append(evt)
            except queue.Empty:
                evt = None

            # q is empty and still have time left, get from q with timeout
            if evt is None:
                timeout = SINK_BATCH_MAX_IDLE_SECS - (time.perf_counter() - last_flush_time)
                if timeout > 0:
                    try:
                        evt = self.q.get(block=True, timeout=timeout)
                        if isinstance(evt, TerminateEvent):
                            continue
                        else:
                            buff.append(evt)
                    except queue.Empty:
                        evt = None

            # here msg is None means timeout
            # timeout or reached max size, trigger flush
            if evt is None or len(buff) >= SINK_BATCH_MAX_SIZE:
                if buff:
                    self._flush(buff)
                    buff = []
                last_flush_time = time.perf_counter()

        # terminated, flush all
        while True:
            try:
                evt = self.q.get_nowait()
                if isinstance(evt, TerminateEvent):
                    continue
                else:
                    buff.append(evt)
            except queue.Empty:
                break
        if buff:
            self._flush(buff)


    def _flush(self, evts: List[Event]):
        self._require_eventhub_client()
        batch = self.client.create_batch()  # type: ignore
        count = 0    # current batch count
        cursor = 0
        while cursor < len(evts):
            # insert batch, until full
            msg = evts[cursor].serialize_json()
            try:
                batch.add(EventData(msg))
                count += 1
                cursor += 1
                continue
            except ValueError:
                pass

            # unexpected error
            if count == 0:
                raise Exception('unexpected error in EventHubSink: batch size is 0 while exceed max bytes')
            
            # send with retries, duration 100ms
            self._send_batch_with_retry(batch, 2, 0.1)

            # send successfully
            self._require_eventhub_client()
            batch = self.client.create_batch()  # type: ignore
            count = 0
            
        if count > 0:
            self._send_batch_with_retry(batch, 2, 0.1)
        
        
    def _send_batch_with_retry(self, batch: EventDataBatch, retry: int, duration: float):
        for retry_id in range(1, retry + 1):
            try:
                self._require_eventhub_client()
                self.client.send_batch(batch)  # type: ignore
                return
            except Exception:
                self._reset_eventhub_client()
                if retry_id == retry:
                    raise
                else:
                    time.sleep(duration)



import os
import threading
from typing import Optional
import uuid

from typeguard import typechecked
from azure.identity import ManagedIdentityCredential

from .eventhub import EventHubSink
from .log import EventLogger
from .logimp import AsyncEventLogger, StdoutSink

__all__ = [
    "infof", 
    "infocf", 
    "warnf", 
    "warncf", 
    "errorf", 
    "errorcf", 
    "fatalf",
    "fatalcf",
    "event",
    "initialize",
]

# create logger
_logger: EventLogger = AsyncEventLogger()


# logger interface
infof = _logger.infof
infocf = _logger.infocf
warnf = _logger.warnf
warncf = _logger.warncf
errorf = _logger.errorf
errorcf = _logger.errorcf
fatalf = _logger.fatalf
fatalcf = _logger.fatalcf
event = _logger.event


# logger init
_logger_init_lock = threading.Lock()
_logger_initialized = False

@typechecked
def initialize(
    eh_hostname: Optional[str] = None,
    client_id: Optional[str] = None,
    eh_conn_str: Optional[str] = None,
    eh_structured: Optional[str] = None,
    eh_unstructured: Optional[str] = None,
    role: Optional[str] = None,
    instance: Optional[str] = None,
):
        '''
        Args:
            eh_hostname: Fully Qualified Namespace aka EH Endpoint URL (*.servicebus.windows.net). Default, read $EVENTHUB_NAMESPACE
            client_id: client_id of service principal. Default, read $UAI_CLIENT_ID
            eh_conn_str: connection string of eventhub namespace. Default, read $EVENTHUB_CONN_STRING
            eh_structured: structured eventhub name. Default, read $EVENTHUB_AUX_STRUCTURED
            eh_unstructured: unstructured eventhub name. Default, read $EVENTHUB_AUX_UNSTRUCTURED
            role: role, Default: RemoteModel_${ENDPOINT_NAME}
            instance: instance, Default: "${ENDPOINT_VERSION}|{os.uname()[1]}" or "${ENDPOINT_VERSION}|{_probably_unique_id()}"
        
        Note: 
            1. either (eh_hostname, client_id, eh_structured or eh_unstructured) 
               or (eh_conn_str, eh_structured or eh_unstructured) is provided, 
               event hub sink will be added
        '''
        global _logger_initialized
        with _logger_init_lock:
            # skip
            if _logger_initialized:
                return
        
            # set default value
            if eh_hostname is None:
                eh_hostname = os.getenv('EVENTHUB_NAMESPACE')
            if client_id is None:
                client_id = os.getenv('UAI_CLIENT_ID')
            if eh_conn_str is None:
                eh_conn_str = os.getenv('EVENTHUB_CONN_STRING')
            if eh_structured is None:
                eh_structured = os.getenv('EVENTHUB_AUX_STRUCTURED')
            if eh_unstructured is None:
                eh_unstructured = os.getenv('EVENTHUB_AUX_UNSTRUCTURED')
            if role is None:
                endpoint_name = os.getenv('ENDPOINT_NAME', 'unknown')
                role = f'RemoteModel_{endpoint_name}'
            if instance is None:
                endpoint_version = os.getenv('ENDPOINT_VERSION', 'unknown')
                instance = f'{endpoint_version}|{os.uname()[1]}'
                if len(instance) > 60:
                    instance = f'{endpoint_version}|{_probably_unique_id()}'
            
            # set variables
            _logger.role = role
            _logger.instance = instance
            
            # event hub sink
            if eh_conn_str:
                if eh_structured:
                    structured_sink = EventHubSink(conn_str=eh_conn_str, name=eh_structured)
                    _logger.add_sink_structured(structured_sink)
                if eh_unstructured:
                    unstructured_sink = EventHubSink(conn_str=eh_conn_str, name=eh_unstructured)
                    _logger.add_sink_unstructured(unstructured_sink)
                
            elif eh_hostname and client_id:
                credential = ManagedIdentityCredential(client_id=client_id)
                if eh_structured:
                    structured_sink = EventHubSink(hostname=eh_hostname, credential=credential, name=eh_structured)
                    _logger.add_sink_structured(structured_sink)
                if eh_unstructured:
                    unstructured_sink = EventHubSink(hostname=eh_hostname, credential=credential, name=eh_unstructured)
                    _logger.add_sink_unstructured(unstructured_sink)
                
            else:
                print('WARNING: logger eventhub sink is disabled')
        
            # stdout sink
            _logger.add_sink_structured(StdoutSink())
            _logger.add_sink_unstructured(StdoutSink())

            # start
            _logger.start()
            _logger_initialized = True
        
                
def _probably_unique_id() -> str:
	u = str(uuid.uuid4())
	return "%s-%s%s" % (u[0:5], u[5:8], u[9:11])




import queue
import time
import traceback
from typing import Any, List, Optional
from dataclasses import dataclass
from threading import Thread, Condition, Event

from typeguard import typechecked
from .base import BaseModel


class ItemFuture:
    def __init__(self, cancellation_token: Optional[Event]=None):
        self._condition = Condition()
        self._flag_done = False
        self._flag_cancelled = False
        self._external_ct = cancellation_token
        self._exception: Optional[Exception] = None
        self._result: Any = None
    
    def get_result(self, timeout: Optional[float]=None) -> Any:
        ''' Return the result of task future
            Args:
                timeout: the number of secodns to wait for result
                         raise PredictTimeoutError when run out of time
                         Default: None, will never timeout
            Raises:
                PredictTimeoutError: If the future is not done before timeout
                PredictCancelledError: If the future is cancelled
        '''
        with self._condition:
            if self._flag_cancelled:
                raise PredictCancelledError('predict cancelled')
            if not self._flag_done:
                self._condition.wait(timeout)
                if self._flag_cancelled:
                    raise PredictCancelledError('predict cancelled')
                if not self._flag_done:
                    raise PredictTimeoutError('predict timeout')
            # from here, _flag_done == True
            if self._exception:
                raise self._exception
            return self._result

    def set_cancelled(self):
        ''' Try to mark cancelled for the task future. Won't actuallly 
            interrupt the process if the task has already started running
        '''
        with self._condition:
            if self._flag_done or self._flag_cancelled:
                return
            self._flag_cancelled = True
            self._condition.notify_all()
            
    def is_cancelled(self) -> bool:
        ''' Return if future is cancelled
        '''
        with self._condition:
            return self._flag_cancelled
        
    def is_external_ct_set(self) -> bool:
        ''' Return if external cancallation token is set
        '''
        return self._external_ct is not None and self._external_ct.is_set()

    def set_result(self, result: Any):
        ''' Set result for the future and mark it done.
            Should only be used by Promise holder
        '''
        with self._condition:
            if self._flag_done:
                return
            self._result = result
            self._flag_done = True
            self._condition.notify_all()
    
    def set_excepted(self, e: Exception):
        ''' Set exception for the future and mark it done.
            Should only be used by Promise holder
        ''' 
        with self._condition:
            if self._flag_done:
                return
            self._exception = e
            self._flag_done = True
            self._condition.notify_all()


@dataclass
class ItemMessage:
    future: ItemFuture
    item: Any


class PredictCancelledError(Exception):
    pass


class PredictTimeoutError(Exception):
    pass


class DynamicBatchModel:
    ''' Wrapped Model for dynamic batching automatically.
    
    A backend thread will serve all the callers. It merges data 
    inputs from multiple calls, and process batchly. Then return 
    corresponding results to each callers.
    
    '''
    
    @typechecked
    def __init__(
        self, 
        model: BaseModel, 
        max_batch_size: int = 12,
        idle_batch_size: int = 5,
        max_batch_interval: float = 0.002,
    ):
        '''
        Args: 
            model: Subclass of BaseMode, to provide business logic
            max_batch_size: Max size of each processing batch
            idle_batch_size: If there's no more data in queue, a new batch will be launched when size reaches this value
            max_batch_interval: Max interval in seconds to wait for items. When waiting time exceeds, will launch a batch immediately.

        '''
        assert idle_batch_size <= max_batch_size

        self.model = model
        self.max_batch_size = max_batch_size
        self.idle_batch_size = idle_batch_size
        self.max_batch_interval = max_batch_interval
        
        self.alive = True
        self.q: queue.Queue[ItemMessage] = queue.Queue()
        self.worker = Thread(target=self._worker_run, daemon=True)
        self.worker.start()
        
    
    @typechecked
    def predict(
        self,
        items: List[Any],
        timeout: Optional[float] = 1.0,
        raise_timeout: bool = True,
        cancellation_token: Optional[Event] = None,
    ) -> List[Any]:
        ''' Predict with dynamic batching
        
        Args:
            items: model input list
            timeout: A floating point number specifying a timeout for the operation in seconds, or None 
                    indicating to wait without timeout
            raise_timeout: When timeout, True: raise PredictTimeoutError; False: set to None for 
                    no result items and return.
            cancellation_token: A event for cancellation. If the event is set, will try to cancel and skip
                    the calculation in downstream batch process. Default: None.
            
        Returns:
            predict output list
        
        '''
        # get futures & put messsage into queue
        futurelist: List[ItemFuture] = []
        for item in items:
            future = ItemFuture(cancellation_token)
            msg = ItemMessage(future, item)
            self.q.put_nowait(msg)
            futurelist.append(future)
        
        # get results
        rs = []
        start_time = time.perf_counter()
        for future in futurelist:
            if timeout is None:
                time_left = None
            else:
                time_left = timeout - (time.perf_counter() - start_time)
            try:
                result = future.get_result(timeout=time_left)
            except PredictTimeoutError:
                future.set_cancelled()
                if raise_timeout:
                    raise
                else:
                    result = None
            rs.append(result)
        return rs


    @typechecked
    def predict_one(
        self,
        item: Any,
        timeout: Optional[float] = 1.0,
        raise_timeout: bool = True,
        cancellation_token: Optional[Event] = None,
    ) -> Any:
        ''' Predict for a single item, with dynamic batching
        
        Args:
            item: model input, a single item
            timeout: A floating point number specifying a timeout for the operation in seconds, or None 
                    indicating to wait without timeout
            raise_timeout: When timeout, True: raise PredictTimeoutError; False: set to None for 
                    no result items and return.
            cancellation_token: A event for cancellation. If the event is set, will try to cancel and skip
                    the calculation in downstream batch process. Default: None.
            
        Returns:
            single predict output
        
        '''
        rs = self.predict(
            [item], 
            timeout=timeout, 
            raise_timeout=raise_timeout, 
            cancellation_token=cancellation_token,
        )
        return rs[0]

    
    def close(self):
        self.alive = False
        
        
    def __del__(self):
        self.close()
        
        
    def _worker_run(self):
        ''' If _worker_inner failed somehow, sleep some time and retry
        '''
        while self.alive:
            try:
                self._worker_inner()
            except Exception:
                # unexpected error
                traceback.print_exc()
                time.sleep(0.1)
                
                
    def _worker_inner(self):
        ''' condition to trigger a batch run:
            1. [batch size] >= self.max_batch_size
            2. queue is empty && [batch size] >= self.idle_batch_size
            3. [now] - [last run time] >= self.max_batch_interval
        '''
        batch: List[ItemMessage] = []          # cached batch items
        last_run_time = time.perf_counter()    # last batch run time
        
        while self.alive:
            # get from q without block & check empty
            try:
                msg = self.q.get_nowait()
                # handle cancellation
                if msg.future.is_external_ct_set():
                    msg.future.set_cancelled()
                    continue
                elif msg.future.is_cancelled():
                    continue
                else:
                    batch.append(msg)
            except queue.Empty:
                msg = None

            # here msg is None means q is empty, idle
            # idle && size >= idle_batch_size, trigger batch run
            if msg is None and len(batch) >= self.idle_batch_size:
                self._worker_run_batch(batch)
                batch = []
                last_run_time = time.perf_counter()
                continue

            # q is empty and still have time left, get from q with timeout
            if msg is None:
                timeout = self.max_batch_interval - (time.perf_counter() - last_run_time)
                if timeout > 0:
                    try:
                        msg = self.q.get(block=True, timeout=timeout)
                        # handle cancellation
                        if msg.future.is_external_ct_set():
                            msg.future.set_cancelled()
                            continue
                        if msg.future.is_cancelled():
                            continue
                        else:
                            batch.append(msg)
                    except queue.Empty:
                        msg = None

            # here msg is None means timeout
            # timeout or batch reached max size, trigger batch run
            if msg is None or len(batch) >= self.max_batch_size:
                if batch:
                    self._worker_run_batch(batch)
                    batch = []
                last_run_time = time.perf_counter()
            
            
    def _worker_run_batch(self, batch: List[ItemMessage]):
        # preprocess & predict
        try:
            input = [msg.item for msg in batch]
            output_preprocess = self.model.preprocess(input)
            output_predict = self.model.predict(output_preprocess)
            
            if len(output_predict) != len(batch):
                raise Exception(f'model output size is {len(output_predict)} while input size is {len(batch)}')
            
        except Exception as e:
            for msg in batch:
                msg.future.set_excepted(e)
            return
        
        # set result for each
        for msg, result in zip(batch, output_predict):
            msg.future.set_result(result)




from abc import ABC, abstractmethod
from typing import Any, List


class BaseModel(ABC):
    '''Base class for models to inherit from
    '''
    
    @abstractmethod
    def predict(self, items: List[Any]) -> List[Any]:
        '''Model predict function
        
        Abstract method which need to be implemented in subclass. It receives
        a bunch of input items, and output same number of predict results.
        
        Args:
            items: Input data items
            
        Returns:
            Predict results for each input data item
        '''
        pass
        
        
    def preprocess(self, items: List[Any]) -> List[Any]:
        '''Preprecossing steps to convert raw data to form consumed by model
        
        Precessing logic is suggested to put in this function. In this way metrics like counting
        and duration will be handled correctly for preprecessong specific. 
        This is not necessary. Default implementation is enought if you just want to
        put every thing in predict function.
        
        Args:
            items: Input data items
            
        Returns:
            Preprocessing results for each input data item
        '''
        return items
    



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pyraisdk', 'pyraisdk.dynbatch', 'pyraisdk.rlog']

package_data = \
{'': ['*']}

install_requires = \
['azure-eventhub>=5.10.1,<6.0.0',
 'azure-identity>=1.7.0,<2.0.0',
 'dataclasses-json>=0.5.7,<0.6.0',
 'typeguard>=2.13.3,<3.0.0']

setup_kwargs = {
    'name': 'sdk-tmp-1207',
    'version': '0.1.1',
    'description': '',
    'long_description': '# pyraisdk\n\n## Dynamic Batching\n\n### Description\nWhen we deploy a model in AML with GPU instance to provide inference service, if it occupies GPU for inferencing in each request separately, it will be quite inefficient. This is a shared module helping to collect data items from different requests, and inferencing batchly in a backend thread. This will considerably improve usage efficiency of GPU.\n\n### Usage Examples\n\nBuild `YourModel` class inherited from `pyraisdk.dynbatch.BaseModel`.\n\n```python\nfrom typing import List\nfrom pyraisdk.dynbatch import BaseModel\n\nclass SimpleModel(BaseModel):\n    def predict(self, items: List[str]) -> List[int]:\n        rs = []\n        for item in items:\n            rs.append(len(item))\n        return rs\n            \n    def preprocess(self, items: List[str]) -> List[str]:\n        rs = []\n        for item in items:\n            rs.append(f\'[{item}]\')\n        return rs\n```\n\nInitialize a `pyraisdk.dynbatch.DynamicBatchModel` with `YourModel` instance, and call `predict / predict_one` for inferencing.\n\n```python\nfrom pyraisdk.dynbatch import DynamicBatchModel\n\n# prepare model\nsimple_model = SimpleModel()\nbatch_model = DynamicBatchModel(simple_model)\n\n# predict\nitems = [\'abc\', \'123456\', \'xyzcccffaffaaa\']\npredictions = batch_model.predict(items)\nassert predictions == [5, 8, 16]\n\n# predict_one\nitem = \'abc\'\nprediction = batch_model.predict_one(item)\nassert prediction == 5\n```\n\nConcurrent requests to `predict / predict_one`, in different threads.\n\n```python\nfrom threading import Thread\nfrom pyraisdk.dynbatch import DynamicBatchModel\n\n# prepare model\nsimple_model = SimpleModel()\nbatch_model = DynamicBatchModel(simple_model)\n\n# thread run function\ndef run(name, num):\n    for step in range(num):\n        item = f\'{name}-{step}\'\n        prediction = batch_model.predict_one(item)\n        assert prediction == len(item) + 2\n\n# start concurrent inference\nthreads = [Thread(target=run, args=(f\'{tid}\', 100)) for tid in range(20)]\nfor t in threads:\n    t.start()\nfor t in threads:\n    t.join()\n```\n\n\n## Loging & Events\n\n### Description\nThis module is for logging and event tracing.\n\n### interface\n\n```python\n\ndef initialize(\n    eh_hostname: Optional[str] = None,\n    client_id: Optional[str] = None,\n    eh_conn_str: Optional[str] = None,\n    eh_structured: Optional[str] = None,\n    eh_unstructured: Optional[str] = None,\n    role: Optional[str] = None,\n    instance: Optional[str] = None,\n)\n```\n\nParameter description for `initialize`:\n- **eh_hostname**: Fully Qualified Namespace aka EH Endpoint URL (*.servicebus.windows.net). Default, read $EVENTHUB_NAMESPACE\n- **client_id**: client_id of service principal. Default, read $UAI_CLIENT_ID\n- **eh_conn_str**: connection string of eventhub namespace. Default, read $EVENTHUB_CONN_STRING\n- **eh_structured**: structured eventhub name. Default, read $EVENTHUB_AUX_STRUCTURED\n- **eh_unstructured**: unstructured eventhub name. Default, read $EVENTHUB_AUX_UNSTRUCTURED\n- **role**: role, Default: RemoteModel_${ENDPOINT_NAME}\n- **instance**: instance, Default: "${ENDPOINT_VERSION}|{os.uname()[1]}" or "${ENDPOINT_VERSION}|{_probably_unique_id()}"\n\n\n```python\n\ndef event(self, key: str, code: str, numeric: float, detail: str=\'\', corr_id: str=\'\', elem: int=-1)\ndef infof(self, format: str, *args: Any)\ndef infocf(self, corr_id: str, elem: int, format: str, *args: Any)\ndef warnf(self, format: str, *args: Any)\ndef warncf(self, corr_id: str, elem: int, format: str, *args: Any)\ndef errorf(self, format: str, *args: Any)\ndef errorcf(self, corr_id: str, elem: int, ex: Optional[Exception], format: str, *args: Any)\ndef fatalf(self, format: str, *args: Any)\ndef fatalcf(self, corr_id: str, elem: int, ex: Optional[Exception], format: str, *args: Any)\n\n```\n\n### examples\n\n```python\n# export EVENTHUB_AUX_UNSTRUCTURED=\'ehunstruct\'\n# export EVENTHUB_AUX_STRUCTURED=\'ehstruct\'\n# export UAI_CLIENT_ID=\'xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx\'\n# export EVENTHUB_NAMESPACE=\'xxx.servicebus.windows.net\'\n\nfrom pyraisdk import rlog\nrlog.initialize()\n\nrlog.infof(\'this is a info message %s\', 123)\nrlog.event(\'LifetimeEvent\', \'STOP_GRACEFUL_SIGNAL\', 0, \'detail info\')\n\n```\n\n```python\n# export EVENTHUB_AUX_UNSTRUCTURED=\'ehunstruct\'\n# export EVENTHUB_AUX_STRUCTURED=\'ehstruct\'\n# export EVENTHUB_CONN_STRING=\'<connection string>\'\n\nfrom pyraisdk import rlog\nrlog.initialize()\n\nrlog.infocf(\'corrid\', -1, \'this is a info message: %s\', 123)\nrlog.event(\'RequestDuration\', \'200\', 0.01, \'this is duration in seconds\')\n\n```\n\n```python\nfrom pyraisdk import rlog\nrlog.initialize(eh_structured=\'ehstruct\', eh_unstructured=\'ehunstruct\', eh_conn_str=\'<eventhub-conn-str>\')\n\nrlog.errorcf(\'corrid\', -1, Exception(\'error msg\'), \'error message: %s %s\', 1,2)\nrlog.event(\'CpuUsage\', \'\', 0.314, detail=\'cpu usage\', corr_id=\'corrid\', elem=-1)\n\n```\n',
    'author': 'Xiaodong Yang',
    'author_email': 'xiaoyan@microsoft.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
